import logging
from django.apps import AppConfig
from django.db import connections
import threading
import asyncio
import os
import sys
from asgiref.sync import sync_to_async


logger = logging.getLogger(__name__)


class ApiConfig(AppConfig):
    name = "api"
    default_auto_field = "django.db.models.BigAutoField"

    @sync_to_async
    def _set_wal_mode(self):
        """Sets the SQLite journal mode to WAL."""
        conn = connections["default"]
        if conn.vendor == "sqlite":
            try:
                # Run PRAGMA within sync_to_async context
                conn.cursor().execute("PRAGMA journal_mode=WAL;")
                logger.info("SQLite journal mode set to WAL.")
            except Exception as e:
                logger.error(f"Failed to set WAL mode: {e}", exc_info=True)

    def ready(self):
        """
        Initialize application components when Django starts.
        This will start the socket and network monitoring services,
        each in a separate thread with its own event loop.
        """
        # Set WAL mode: Use create_task if loop running, else run synchronously
        try:
            loop = asyncio.get_running_loop()
            logger.debug("Event loop running, scheduling WAL mode set via create_task.")
            loop.create_task(self._set_wal_mode())
        except RuntimeError:
            logger.debug("No event loop running, setting WAL mode synchronously.")
            conn = connections["default"]
            if conn.vendor == "sqlite":
                try:
                    conn.cursor().execute("PRAGMA journal_mode=WAL;")
                    logger.info("SQLite journal mode set to WAL synchronously.")
                except Exception as e:
                    logger.error(
                        f"Failed to set WAL mode synchronously: {e}", exc_info=True
                    )

        logger.debug("API AppConfig ready method called.")
        logger.debug("sys.argv = %s", sys.argv)

        is_server_command = any(
            word in arg
            for arg in sys.argv
            for word in ("runserver", "uvicorn", "gunicorn", "daphne", "hypercorn")
        )

        is_main_process = (
            os.environ.get("RUN_MAIN") == "true" or "uvicorn" in sys.argv[0]
        )

        if not is_server_command or not is_main_process:
            logger.debug(
                "Skipping services (server=%s, main=%s)",
                is_server_command,
                is_main_process,
            )
            return

        logger.info("Initializing API application background services...")

        # Start Socket Monitor
        socket_thread = threading.Thread(
            target=self._start_socket_monitor, daemon=True
        )
        socket_thread.start()
        logger.info("Socket monitor thread initiated.")

        # # Start Network Monitor
        # network_thread = threading.Thread(
        #     target=self._start_network_monitor, daemon=True
        # )
        # network_thread.start()
        # logger.info("Network monitor thread initiated.")

    def _start_socket_monitor(self):
        """
        Starts the SocketMonitorService in a dedicated thread
        with its own event loop.
        """
        try:
            from .services import SocketMonitorService

            logger.info("Starting Socket Monitor thread...")
            loop = asyncio.new_event_loop()

            # Set the event loop for this thread
            asyncio.set_event_loop(loop)

            # Get monitor instance and start monitoring
            monitor = SocketMonitorService.get_instance()

            try:
                loop.run_until_complete(monitor.start_monitoring())
                logger.info("Socket Monitor started successfully.")
                # Keep the loop running to process socket events
                loop.run_forever()
            except Exception as e:
                logger.error(
                    f"Error in Socket Monitor loop: {e}", exc_info=True
                )
                sys.exit(1)
            finally:
                loop.close()
                logger.info("Socket Monitor loop closed.")

        except Exception as e:
            logger.error(f"Error starting Socket Monitor: {e}", exc_info=True)
            sys.exit(1)

    def _start_network_monitor(self):
        """
        Starts the NetworkMonitor in a dedicated thread with its own event loop.
        """
        try:
            from .services import NetworkMonitor

            logger.info("Starting Network Monitor thread...")
            loop = asyncio.new_event_loop()

            # Set the event loop for this thread
            asyncio.set_event_loop(loop)

            # Get network monitor instance and start monitoring
            monitor = NetworkMonitor.get_instance()

            try:
                loop.run_until_complete(monitor.start_monitoring())
                # Keep the loop running to process network events
                loop.run_forever()
            except Exception as e:
                logger.error(
                    f"Error in Network Monitor loop: {e}", exc_info=True
                )
            finally:
                loop.close()
                logger.info("Network Monitor loop closed.")

        except Exception as e:
            logger.error(
                f"Error in Network Monitor thread: {e}", exc_info=True
            )
