import os
import logging
import pty
import fcntl
import struct
import termios
import asyncio
import subprocess
import re
import json
import select
from channels.generic.websocket import AsyncWebsocketConsumer
from asgiref.sync import sync_to_async

# Configure logging
logger = logging.getLogger(__name__)


class StatsConsumer(AsyncWebsocketConsumer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.process = None
        self.master_fd = None
        self.stream_task = None
        self.keep_running = False

        # Hardcoded reference uptime (used as zero)
        self.reference_uptime = "15102d-43:25:55"
        self.reference_seconds = self.parse_uptime_to_seconds(self.reference_uptime)
        logger.info(
            f"Reference uptime set to {self.reference_uptime} ({self.reference_seconds} seconds)"
        )

    async def connect(self):
        await self.accept()
        logger.info("WebSocket connection established.")
        self.keep_running = True

        # Create a PTY for KelVPNCli process
        self.master_fd, slave_fd = pty.openpty()

        # Set virtual terminal size (required for correct connection time output)
        desired_width = 200
        desired_height = 24
        fcntl.ioctl(
            slave_fd,
            termios.TIOCSWINSZ,
            struct.pack("HHHH", desired_height, desired_width, 0, 0),
        )

        # Start KelVPNCli -T as a subprocess
        self.process = subprocess.Popen(
            ["/opt/kelvpn/bin/KelVPNCli", "-T"],
            stdout=slave_fd,
            stderr=slave_fd,
            close_fds=True,
        )
        os.close(slave_fd)
        logger.info("KelVPNCli process started successfully.")

        # Set master_fd to non-blocking mode
        flags = fcntl.fcntl(self.master_fd, fcntl.F_GETFL)
        fcntl.fcntl(self.master_fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        # Start background task to read statistics
        self.stream_task = asyncio.create_task(self.stream_stats())

    async def disconnect(self, close_code):
        logger.info(
            f"WebSocket connection closed (code: {close_code}). Stopping KelVPNCli process..."
        )
        self.keep_running = False

        # Cancel background task
        if self.stream_task:
            self.stream_task.cancel()

        # Terminate KelVPNCli process
        if self.process:
            self.process.terminate()
            self.process.wait()
            logger.info("KelVPNCli process terminated.")

        # Close PTY file descriptor
        if self.master_fd is not None:
            try:
                os.close(self.master_fd)
                self.master_fd = None
                logger.info("PTY master file descriptor closed.")
            except OSError as e:
                logger.warning(f"Error closing PTY master file descriptor: {str(e)}")

    async def stream_stats(self):
        """
        Main loop: every second, check for new data from PTY.
        If available, read the data, extract statistics, adjust uptime using
        15102d-43:25:55 as the baseline, and send the corrected stats via WebSocket.
        """
        try:
            while self.keep_running:
                await asyncio.sleep(1)
                ready, _, _ = select.select([self.master_fd], [], [], 0)
                if not ready:
                    continue

                try:
                    data = os.read(self.master_fd, 4096)
                except OSError:
                    continue
                if not data:
                    break

                decoded_data = data.decode("utf-8", errors="replace")
                lines = decoded_data.splitlines()

                for line in lines:
                    line = line.strip()
                    if not line:
                        continue

                    if "State: Disconnect" in line:
                        logger.warning(
                            "KelVPN state is 'Disconnect'. Updating config and closing WebSocket."
                        )
                        try:
                            await self.mark_config_disconnected()
                        except Exception as e:
                            logger.error(
                                f"Failed to update config for 'Disconnect' state: {e}"
                            )

                        try:
                            await self.send(
                                text_data=json.dumps(
                                    {
                                        "event": "disconnect",
                                        "message": "KelVPN is disconnected",
                                    }
                                )
                            )
                        except Exception as e:
                            logger.error(
                                f"Error sending disconnect event to WebSocket: {str(e)}"
                            )

                        self.keep_running = False
                        break
                    if "uptime" not in line.lower():
                        continue

                    logger.debug(f"Processing line with uptime: {line}")
                    stats = self.parse_stats(line)
                    if stats:
                        try:
                            await self.send(text_data=json.dumps(stats))
                            logger.info(f"Sent stats: {stats}")
                        except Exception as e:
                            logger.error(f"Error sending WebSocket data: {str(e)}")
                            self.keep_running = False
                            break

        except asyncio.CancelledError:
            logger.info("Stream stats task was cancelled.")
        except Exception as e:
            logger.error(f"Unexpected error in stream_stats: {str(e)}")

    def parse_stats(self, line):
        """
        Extract key statistics from the received line.
        """
        download_match = re.search(r"download\s+([\d.]+\s+\w+)", line, re.IGNORECASE)
        upload_match = re.search(r"upload\s+([\d.]+\s+\w+)", line, re.IGNORECASE)
        received_match = re.search(
            r"([\d.]+\s+\w+)\s+bytes received", line, re.IGNORECASE
        )
        sent_match = re.search(r"([\d.]+\s+\w+)\s+bytes sent", line, re.IGNORECASE)
        packets_received_match = re.search(
            r"(\d+)\s+packets received", line, re.IGNORECASE
        )
        packets_sent_match = re.search(r"(\d+)\s+packets sent", line, re.IGNORECASE)
        uptime_match = re.search(r"((?:\d+d-)?[\d:]+)\s+uptime", line, re.IGNORECASE)

        if not all(
            [
                download_match,
                upload_match,
                received_match,
                sent_match,
                packets_received_match,
                packets_sent_match,
                uptime_match,
            ]
        ):
            return None

        raw_uptime = uptime_match.group(1)
        corrected_uptime = self.recalculate_uptime(raw_uptime)

        return {
            "download": download_match.group(1),
            "upload": upload_match.group(1),
            "bytes_received": received_match.group(1),
            "bytes_sent": sent_match.group(1),
            "packets_received": int(packets_received_match.group(1)),
            "packets_sent": int(packets_sent_match.group(1)),
            "uptime": corrected_uptime,
        }

    def recalculate_uptime(self, uptime_str: str) -> str:
        """
        Convert uptime string to seconds, subtract hardcoded reference time,
        and format the difference as HH:MM:SS.
        """
        current_seconds = self.parse_uptime_to_seconds(uptime_str)
        diff = max(0, current_seconds - self.reference_seconds)

        # Format difference as HH:MM:SS
        hh = diff // 3600
        mm = (diff % 3600) // 60
        ss = diff % 60
        return f"{hh:02}:{mm:02}:{ss:02}"

    def parse_uptime_to_seconds(self, uptime_str: str) -> int:
        """
        Convert an uptime string (e.g., "15102d-43:25:55" or "HH:MM:SS") into total seconds.
        Returns 0 if parsing fails.
        """
        days = 0
        time_part = uptime_str

        if "d-" in uptime_str:
            day_part, time_part = uptime_str.split("d-")
            try:
                days = int(day_part)
            except ValueError:
                days = 0

        parts = time_part.split(":")
        if len(parts) == 3:
            h, m, s = parts
        elif len(parts) == 2:
            h, m = parts
            s = "0"
        else:
            logger.warning(f"Unexpected uptime format: {uptime_str}")
            return 0

        try:
            h = int(h)
            m = int(m)
            s = int(s)
        except ValueError:
            logger.warning(f"Invalid uptime string: {uptime_str}")
            return 0

        total_seconds = days * 86400 + h * 3600 + m * 60 + s
        return total_seconds

    @sync_to_async
    def mark_config_disconnected(self):
        from api.models import Config

        config_path = "/opt/kelvpn/etc/KelVPNCheck.conf"

        try:
            config = Config.objects.first()
            if config:
                config.is_connected = False
                config.save()
                logger.info("Config updated: is_connected=False (disconnect detected)")
            else:
                logger.warning("No Config object found to mark as disconnected.")
        except Exception as e:
            logger.error(f"Failed to update Config model: {e}")

        try:
            with open(config_path, "r") as file:
                lines = file.readlines()

            new_lines = []
            found_connected = False

            for line in lines:
                if line.startswith("KELVPN_CONNECTED="):
                    new_lines.append("KELVPN_CONNECTED=false\n")
                    found_connected = True
                else:
                    new_lines.append(line)

            if not found_connected:
                new_lines.append("KELVPN_CONNECTED=false\n")

            with open(config_path, "w") as file:
                file.writelines(new_lines)

            logger.info("KelVPNCheck.conf updated: KELVPN_CONNECTED=false")

        except Exception as e:
            logger.error(f"Failed to update KelVPNCheck.conf: {e}")
