from django.db import models
from django.utils import timezone
from django.conf import settings


class SerialHistory(models.Model):
    serial_key = models.CharField(max_length=20, unique=True)
    updated_at = models.DateTimeField(auto_now=True)
    status = models.CharField(max_length=20, default="active")
    remaining_days = models.IntegerField(default=30)
    expired_date = models.DateTimeField(null=True, blank=True)

    def __str__(self):
        return self.serial_key


class Server(models.Model):
    server_id = models.CharField(max_length=2, unique=True)
    name = models.CharField(max_length=100)
    ping = models.IntegerField(default=-1)
    selected = models.BooleanField(default=False)
    ip_address = models.GenericIPAddressField(null=True, blank=True)

    def __str__(self):
        return f"{self.server_id} - {self.name}"


class Config(models.Model):
    ip = models.GenericIPAddressField()
    location = models.CharField(max_length=100)
    dns = models.GenericIPAddressField()
    gateway = models.GenericIPAddressField()
    remaining_days = models.IntegerField(default=180)
    selected_server = models.ForeignKey(
        Server, on_delete=models.SET_NULL, null=True, blank=True
    )
    is_activated = models.BooleanField(default=False)
    is_connected = models.BooleanField(default=False)

    session_state = models.CharField(
        max_length=20,
        choices=[
            ("false", "false"),
            ("true", "true"),
            ("false_to_true", "false_to_true"),
            ("true_to_false", "true_to_false"),
        ],
        default="false",
    )
    stream_state = models.CharField(
        max_length=20, choices=session_state.choices, default="false"
    )
    tunnel_state = models.CharField(
        max_length=20, choices=session_state.choices, default="false"
    )

    def __str__(self):
        return f"Config - {self.location}"

    def update_connection_flag(self):
        def truthy(v):
            return "true" in v

        self.is_connected = (
            truthy(self.session_state)
            and truthy(self.stream_state)
            and truthy(self.tunnel_state)
        )
        self.save()


class WiFiStatus(models.Model):
    interface = models.CharField(max_length=10, blank=True, null=True)
    band = models.CharField(max_length=10)  # "2.4GHz", "5GHz", "6GHz"
    ssid = models.CharField(max_length=50, blank=True, null=True)
    channel = models.CharField(max_length=10, blank=True, null=True)
    hw_mode = models.CharField(max_length=20, blank=True, null=True)
    frequency = models.CharField(max_length=10)  # Частота, например, "2412 MHz"
    transmit_power = models.FloatField()  # Мощность передатчика в dBm
    wpa_passphrase = models.CharField(max_length=64, blank=True, null=True)
    enabled = models.BooleanField(default=False)


class WorkProfile(models.Model):
    PROFILE_TYPE_CHOICES = [
        ("no_restriction", "No Restriction"),
        ("selection_mode", "Selection Mode"),
        ("exclusion_mode", "Exclusion Mode"),
    ]

    name = models.CharField(max_length=100, unique=True)
    profile_type = models.CharField(
        max_length=20, choices=PROFILE_TYPE_CHOICES, default="no_restriction"
    )
    is_active = models.BooleanField(default=True)
    selected_urls = models.ManyToManyField(
        "ProfileURL", blank=True, related_name="work_profiles"
    )
    selected_apps = models.ManyToManyField(
        "Application", blank=True, related_name="work_profiles"
    )

    def __str__(self):
        return self.name


class ProfileURL(models.Model):
    url = models.CharField(max_length=255, unique=True)

    def __str__(self):
        return self.url


class Application(models.Model):
    name = models.CharField(max_length=100, unique=True)

    def __str__(self):
        return self.name


class UpdateStatus(models.Model):
    status = models.CharField(max_length=20, default="NOT_STARTED")
    log_file = models.CharField(max_length=255, blank=True, null=True)
    pid = models.IntegerField(blank=True, null=True)
    unit_name = models.CharField(max_length=255, blank=True, null=True)
    started_at = models.DateTimeField(auto_now_add=True)
    completed_at = models.DateTimeField(null=True, blank=True)


class UserTokenBlock(models.Model):
    """
    Stores the timestamp after which all tokens for a user are invalid.
    """
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="token_block",
    )
    blocked_at = models.DateTimeField(default=timezone.now)

    def __str__(self):
        return f"TokenBlock(user={self.user}, blocked_at={self.blocked_at})"
