import asyncio
import json
import logging
import os
import re
import subprocess
import socket
import threading
from datetime import datetime, timezone as dt_timezone
from typing import Any, Dict, List, Optional, Tuple

from asgiref.sync import sync_to_async
from django.conf import settings
from django.db import transaction
from filelock import FileLock, Timeout
from pyroute2 import IPRoute

from .models import Config, Server, SerialHistory, UpdateStatus, WiFiStatus
from .scheme import UpdateCheckResponse, UpdateItem
from .utils import command_handler, register_socket_handlers

logger = logging.getLogger(__name__)
LOCK_PATH = "/tmp/wifi_manager.lock"
LOCK_TIMEOUT = 20


class KelVPNCliService:
    def __init__(self) -> None:
        self.cli_path: str = "/opt/kelvpn/bin/KelVPNCli"
        self.config_path: str = "/opt/kelvpn/share/KelVPNCheck.conf"

    def execute_command(
        self, args: List[str], timeout: int = 90, retry_on_timeout: bool = True
    ) -> str:
        """
        Executes a command with the provided arguments and returns the result.
        If a timeout error occurs, it restarts KelVPNService and retries once.

        :param args: List of command arguments.
        :param timeout: Timeout for the command execution in seconds.
        :param retry_on_timeout: Whether to retry the command execution after a
            timeout.
        :return: Standard output from the command.
        :raises Exception: If the command execution fails after retrying.
        """
        try:
            result = subprocess.run(
                args,
                capture_output=True,
                text=True,
                check=True,
                timeout=timeout,
            )
            output = result.stdout.strip()

            if "The command execution time" in output and retry_on_timeout:
                logger.error(
                    f"Detected timeout error in command output: {output}"
                )
                return self.handle_timeout_and_retry(args, timeout)

            return output

        except subprocess.CalledProcessError as e:
            logger.error(f"Command execution error: {e.stderr.strip()}")
            raise Exception(f"Command execution error: {e.stderr.strip()}") from e
        except subprocess.TimeoutExpired:
            logger.error("Command execution time exceeded the limit.")
            if retry_on_timeout:
                return self.handle_timeout_and_retry(args, timeout)
            raise Exception(
                "Command execution time exceeded the limit."
            ) from None

    def handle_timeout_and_retry(self, args: List[str], timeout: int) -> str:
        """
        Handles command execution timeout by restarting KelVPNService and
        retrying the command.

        :param args: List of command arguments.
        :param timeout: Timeout for the command execution in seconds.
        :return: Standard output from the retried command.
        :raises Exception: If the retry also fails.
        """
        logger.warning("Restarting KelVPNService due to timeout...")

        cmd_restart_kelvpn: List[str] = [
            "systemctl",
            "restart",
            "KelVPNService.service",
        ]
        logger.debug(
            f"Restarting KelVPNService command: {' '.join(cmd_restart_kelvpn)}"
        )

        try:
            kelvpn_output: str = self.execute_command(
                cmd_restart_kelvpn, retry_on_timeout=False
            )
            logger.info(f"KelVPNService restarted successfully: {kelvpn_output}")
        except Exception as e:
            logger.error(f"Error during KelVPNService restart: {e}")
            raise Exception(f"Failed to restart KelVPNService: {e}")

        import time

        logger.debug("Sleeping for 2 seconds after restart...")
        time.sleep(2)

        logger.debug(f"Retrying command: {' '.join(args)}")
        try:
            result = subprocess.run(
                args,
                capture_output=True,
                text=True,
                check=True,
                timeout=timeout,
            )
            output = result.stdout.strip()
            logger.info(f"Retry successful. Command output: {output}")
            return output
        except Exception as e:
            logger.error(f"Retry command execution error: {e}")
            raise Exception(f"Retry command execution error: {e}")

    def start_service(self, server_id: str, server_name: str) -> Tuple[str, bool]:
        """
        Starts KelVPNCli with the selected server.

        :param server_id: The ID of the server to connect to.
        :param server_name: The name of the server.
        :return: Tuple containing the command output and a success flag.
        """
        cmd_connect: List[str] = [self.cli_path, "-C", server_id]
        outputs = []

        logger.debug(
            f"Start service called with server_id: {server_id}, "
            f"server_name: {server_name}"
        )

        # Stop KelVPNCheck.service
        try:
            if self.is_service_active("KelVPNCheck.service"):
                cmd_stop_check: List[str] = [
                    "systemctl",
                    "stop",
                    "KelVPNCheck.service",
                ]
                logger.debug(
                    "Stopping KelVPNCheck.service with command: "
                    f"{' '.join(cmd_stop_check)}"
                )
                check_service_output: str = self.execute_command(cmd_stop_check)
                outputs.append(f"KelVPNCheck.service stopped: {check_service_output}")
                logger.debug(outputs[-1])
            else:
                logger.debug("KelVPNCheck.service is not active; no need to stop it.")
        except Exception as e:
            if "not loaded" in str(e).lower() or "not found" in str(e).lower():
                outputs.append("KelVPNCheck.service is not running or not found.")
                logger.warning(outputs[-1])
            else:
                outputs.append(f"KelVPNCheck.service stop failed: {e}")
                logger.error(outputs[-1])

        # Attempt connection
        try:
            logger.debug(
                f"Attempting to connect using command: {' '.join(cmd_connect)}"
            )
            output: str = self.execute_command(cmd_connect)
            logger.debug(f"Output from connection attempt: {output}")
        except Exception as e:
            logger.error(
                f"Command execution error during connection attempt: {e}"
            )
            return f"Command execution error: {e}", False

        # Case: Already connected
        if "There is already a connection to the server." in output:
            logger.info(
                "Detected: Already connected to the server. "
                "Updating configuration accordingly."
            )
            self.update_config(
                is_connected=True, is_activated=True, server_name=server_name
            )
            try:
                cmd_start_check: List[str] = [
                    "systemctl",
                    "start",
                    "KelVPNCheck.service",
                ]
                logger.debug(
                    "Starting KelVPNCheck.service with command: "
                    f"{' '.join(cmd_start_check)}"
                )
                check_service_output: str = self.execute_command(cmd_start_check)
                logger.info(
                    "KelVPNCheck.service started successfully: "
                    f"{check_service_output}"
                )
            except Exception as e:
                logger.error(
                    "Error starting KelVPNCheck.service after already "
                    f"connected: {e}"
                )
            return output, True

        # Case: Serial key not found in database
        if "Serial key not found in database" in output:
            logger.warning("Serial key not found in database. Deactivating key.")
            self.deactivate_serial()
            return "Serial key not found and has been deactivated.", False

        # Case: General connection failure
        # (output doesn't contain expected markers)
        if "State: Connect" not in output or "Success!" not in output:
            self.update_config(is_connected=False, server_name=server_name)
            logger.error(f"Connection attempt failed. Output: {output}")
            return output, False

        # Start KelVPNCheck.service
        try:
            cmd_start_check: List[str] = [
                "systemctl",
                "start",
                "KelVPNCheck.service",
            ]
            logger.debug(
                "Starting KelVPNCheck.service with command: "
                f"{' '.join(cmd_start_check)}"
            )
            check_service_output: str = self.execute_command(cmd_start_check)
            logger.info(
                "KelVPNCheck.service started successfully: " f"{check_service_output}"
            )
        except Exception as e:
            logger.error(f"Error starting KelVPNCheck.service: {e}")

        # Successful connection
        self.update_config(
            is_connected=True, is_activated=True, server_name=server_name
        )
        logger.info("Connection established successfully. Updating configuration.")

        return output, True

    def disconnect_service(self) -> str:
        """
        Disconnects KelVPNCli and stops related services.

        :return: Combined output from the commands as a string.
        """
        outputs = []
        try:
            # Disconnect KelVPNCli
            cmd_disconnect: List[str] = [self.cli_path, "-D"]
            disconnect_output: str = self.execute_command(cmd_disconnect)
            outputs.append(f"KelVPNCli disconnected: {disconnect_output}")
            logger.info(outputs[-1])
        except Exception as e:
            outputs.append(f"KelVPNCli disconnect failed: {e}")
            logger.error(outputs[-1])

        # Stop KelVPNCheck.service
        try:
            if self.is_service_active("KelVPNCheck.service"):
                cmd_stop_check: List[str] = [
                    "systemctl",
                    "stop",
                    "KelVPNCheck.service",
                ]
                check_service_output: str = self.execute_command(cmd_stop_check)
                outputs.append(f"KelVPNCheck.service stopped: {check_service_output}")
                logger.info(outputs[-1])
            else:
                outputs.append("KelVPNCheck.service is not running or not found.")
                logger.warning(outputs[-1])
        except Exception as e:
            if "not loaded" in str(e).lower() or "not found" in str(e).lower():
                outputs.append("KelVPNCheck.service is not running or not found.")
                logger.warning(outputs[-1])
            else:
                outputs.append(f"KelVPNCheck.service stop failed: {e}")
                logger.error(outputs[-1])

        # # Stop KelVPNService.service
        # try:
        #     if self.is_service_active("KelVPNService.service"):
        #         cmd_stop_service: List[str] = [
        #             "systemctl",
        #             "stop",
        #             "KelVPNService.service",
        #         ]
        #         service_output: str = self.execute_command(cmd_stop_service)
        #         outputs.append(
        #             f"KelVPNService.service stopped: {service_output}"
        #         )
        #         logger.info(outputs[-1])
        #     else:
        #         outputs.append(
        #             "KelVPNService.service is not running or not found."
        #         )
        #         logger.warning(outputs[-1])
        # except Exception as e:
        #     if (
        #         "not loaded" in str(e).lower()
        #         or "not found" in str(e).lower()
        #     ):
        #         outputs.append(
        #             "KelVPNService.service is not running or not found."
        #         )
        #         logger.warning(outputs[-1])
        #     else:
        #         outputs.append(f"KelVPNService.service stop failed: {e}")
        #         logger.error(outputs[-1])

        # Update configuration
        try:
            self.update_config(is_connected=False)
        except Exception as e:
            logger.error(f"Error updating configuration: {e}")

        return "\n".join(outputs)

    def is_service_active(self, service_name: str) -> bool:
        """
        Checks whether the service exists and whether it is active.
        :param service_name: The name of the service.
        :return: True if the service is active, otherwise False.
        """
        try:
            cmd_check: List[str] = [
                "systemctl",
                "list-units",
                "--full",
                "--all",
                service_name,
            ]
            result: str = self.execute_command(cmd_check).strip()
        except Exception as e:
            logger.warning(f"Service check failed for {service_name}: {e}")
            return False

        if service_name in result:
            try:
                active_result = self.execute_command(
                    ["systemctl", "is-active", service_name]
                ).strip()
                return active_result == "active"
            except Exception as e:
                logger.warning(
                    f"Service is-active check failed for {service_name}: {e}"
                )
                return False
        else:
            logger.warning(f"Service {service_name} is not found in systemd.")
            return False

    def restart_service(self) -> Tuple[str, bool]:
        """
        Restarts the KelVPNService and kelvpn_backend.service.

        :return: Tuple containing the output from the restart commands and
                 success flag.
        """
        try:
            # Capture previous connection state and selected server
            prev_cfg = Config.objects.first()
            was_connected = prev_cfg.is_connected if prev_cfg else False
            prev_server = prev_cfg.selected_server if prev_cfg else None

            # Update configuration
            self.update_config(is_connected=False)

            # Restart KelVPNService
            cmd_restart_kelvpn: List[str] = [
                "systemctl",
                "restart",
                "KelVPNService.service",
            ]
            kelvpn_output: str = self.execute_command(cmd_restart_kelvpn)
            logger.info(f"KelVPNService restarted successfully: {kelvpn_output}")

            # Attempt to reconnect to previous server if it was connected
            if was_connected and prev_server:
                try:
                    out, success = self.start_service(
                        prev_server.server_id, prev_server.name
                    )
                    if success:
                        logger.info(
                            f"Auto-reconnected to server {prev_server.name} successfully"
                        )
                    else:
                        logger.warning(
                            f"Auto-reconnect failed: {out}"
                        )
                except Exception as e:
                    logger.error(f"Error during auto-reconnect: {e}")
            return kelvpn_output + "\n", True
        except Exception as e:
            logger.error(f"Error restarting services: {str(e)}")
            return str(e), False

    def reboot_system(self) -> str:
        """
        Reboots the system

        :return: Standard output from the reboot command.
        """
        self.update_config(is_connected=False)
        cmd_reboot: List[str] = ["reboot"]
        output: str = self.execute_command(cmd_reboot)
        logger.info("System reboot initiated.")
        return output

    def get_service_status(self) -> str:
        """
        Retrieves the status of the KelVPNService.

        :return: Standard output from the status command.
        """
        cmd_status: List[str] = [
            "systemctl",
            "status",
            "KelVPNService",
            "--no-pager",
        ]
        output: str = self.execute_command(cmd_status)
        return output

    def set_serial(self, serial_key: str) -> Tuple[str, bool, str]:
        """
        Sets the serial key using the best server and updates the remaining
        days dynamically.

        :param serial_key: The serial key to activate.
        :return: Tuple containing the command output, success flag, and a message
                 indicating the result.
        """
        servers_ping: List[Dict[str, Any]] = self.get_server_pings()
        best_server = min(servers_ping, key=lambda server: server["ping"])
        logger.info(f"best_server {best_server} for activation.")

        if not best_server:
            raise Exception("No server available for activation.")

        server_id = best_server["server_id"]

        logger.info(f"Using best server with ID {server_id} for activation.")

        cmd_set_serial: List[str] = [self.cli_path, "-C", server_id, "-S", serial_key]
        logger.debug(f"Command set serial: {cmd_set_serial}")

        output: str = self.execute_command(cmd_set_serial)
        logger.debug(f"Set serial command output: {output}")

        success, message = self._parse_set_serial_output(output)
        if not success:
            return output, False, message

        # remaining_days = self._get_serial_days_left()
        # if remaining_days <= 0:
        #     logger.error("Failed to retrieve remaining days from serial activation.")
        #     return output, False, "Unable to activate serial key."

        with transaction.atomic():
            config: Optional[Config] = Config.objects.first()

            if not config:
                raise Exception("Configuration not found, cannot proceed.")

            selected_server = Server.objects.filter(server_id=server_id).first()

            if not selected_server:
                raise Exception(
                    f"Selected server with ID {server_id} not found in " f"database."
                )

            config.selected_server = selected_server
            config.is_activated = True
            config.is_connected = True
            config.save()

            logger.info(
                f"Config updated: is_activated=True, is_connected=True, "
                f"selected_server={selected_server.name}"
            )

            # serial, created = SerialHistory.objects.update_or_create(
            #     serial_key=serial_key,
            #     defaults={
            #         "status": "active",
            #         "remaining_days": remaining_days,
            #     },
            # )

            # if created:
            #     logger.info(
            #         f"New serial key {serial.serial_key} created with "
            #         f"{remaining_days} days."
            #     )
            # else:
            #     logger.info(
            #         f"Existing serial key {serial.serial_key} updated with "
            #         f"{remaining_days} days."
            #     )
            #     serial.remaining_days = remaining_days
            #     if serial.status != "active":
            #         serial.status = "active"
            #     serial.save()

        return output, True, message

    def _get_serial_days_left(self) -> int:
        """
        Retrieves the number of remaining days for the current serial key
        using KelVPNCli -AI.

        :return: Number of days left or -1 if parsing failed.
        """
        cmd_auth_info: List[str] = [self.cli_path, "-AI"]
        try:
            output = self.execute_command(cmd_auth_info)
            logger.debug(f"Authorization info output: {output}")

            match = re.search(r"days left: (\d+) days", output)
            if match:
                days_left = int(match.group(1))
                logger.info(f"Days left from serial key: {days_left}")
                return days_left

            logger.error("Failed to parse days left from authorization info.")
            return -1

        except Exception as e:
            logger.error(f"Error retrieving days left from serial key: {str(e)}")
            return -1

    def _parse_set_serial_output(self, output: str) -> Tuple[bool, str]:
        """
        Parses the output of the set_serial command to determine success and
        returns a message.

        :param output: The standard output from the set_serial command.
        :return: Tuple of boolean success flag and message.
        """

        if "State: Connect" in output and "Success!" in output:
            return True, "Serial key activated and connected successfully."

        error_match = re.search(r"Error: code: (-?\d+), message: (.*)", output)
        if error_match:
            code_int = int(error_match.group(1))
            error_msg = error_match.group(2).strip()
            abs_code = abs(code_int)
            # Expired or not found serial key
            if abs_code == 31000:
                if "expired" in error_msg.lower():
                    return False, "Activation failed. The serial key has expired."
                if "not found in database" in error_msg.lower():
                    return False, "Activation failed. The serial key was not found in the database."
            # Activation-specific errors
            if code_int == 104:
                return False, (
                    "Failed to connect to the server. "
                    "The reason for connection failure on the server side."
                )
            if "already activated" in error_msg.lower():
                return False, "Activation failed. The serial key has been activated on another device."
            # Fallback generic message
            return False, "Activation failed. Please try again"

        if "No activated key found" in output and "State: Connecting" in output:
            if "Error: code:" in output:
                error_code = re.search(r"Error: code: (\d+)", output)
                err_msg = (
                    f"Activation failed with error code "
                    f"{error_code.group(1) if error_code else 'unknown'}."
                )
                logger.error(err_msg)
                return False, err_msg

        if "State: Disconnect" in output and "No activated key found" in output:
            logger.error("No activated key found. Connection attempt failed.")
            return False, "No activated key found. Connection attempt failed."

        if "Invalid serial key format" in output:
            logger.error("Invalid serial key format detected.")
            return False, "Invalid serial key format."

        if "Error: code:" in output:
            error_code = re.search(r"Error: code: (\d+)", output)
            message = re.search(r"message: (.+)", output)
            err_msg = (
                f"Error code {error_code.group(1)}, " f"message: {message.group(1)}"
                if error_code and message
                else "Unknown error occurred."
            )
            logger.error(f"Activation failed: {err_msg}")
            return False, err_msg

        logger.error(f"Unexpected output: {output}")
        return False, "Unexpected error occurred. Check logs for details."

    def _format_serial_key(self, serial: str) -> str:
        """
        Formats a serial key into the database format (XXXX-XXXX-XXXX-XXXX).
        If the serial key already contains hyphens or its length is not 16,
        returns it unchanged.
        """
        if not serial:
            return ""
        serial = serial.replace("-", "")
        if len(serial) != 16:
            return serial
        return f"{serial[0:4]}-{serial[4:8]}-{serial[8:12]}-{serial[12:16]}"

    def get_serial_history(self) -> List[Dict[str, Any]]:
        """
        Retrieves and stores the serial history.
        Additionally, checks the authorization info (-AI) and, if the days left
        for the serial key are negative, marks that serial as expired and updates
        the configuration (is_activated=False).

        :return: List of dictionaries containing serial keys and their creation
                 timestamps.
        """
        cmd_history: List[str] = [self.cli_path, "-H"]
        output: str = self.execute_command(cmd_history)
        serial_keys: List[str] = self._parse_serial_history(output)

        with transaction.atomic():
            for key in serial_keys:
                try:
                    obj, created = SerialHistory.objects.get_or_create(
                        serial_key=key, defaults={"status": "deactivated"}
                    )
                    if created:
                        logger.info(f"Created new SerialHistory for serial key {key}.")
                    else:
                        logger.info(f"Serial key {key} already exists.")
                except Exception as e:
                    logger.error(
                        "Error while creating or getting SerialHistory for "
                        f"serial key {key}: {str(e)}"
                    )

        # Retrieve authorization info using the -AI command.
        cmd_auth_info: List[str] = [self.cli_path, "-AI"]
        try:
            auth_output: str = self.execute_command(cmd_auth_info)
        except Exception as e:
            logger.error(f"Error executing -AI command: {e}")
            auth_output = ""

        logger.debug(f"Authorization info output: {auth_output}")

        serial_key_match = re.search(
            r"serial key:\s*([A-Z0-9]+)", auth_output, re.IGNORECASE
        )
        days_left_match = re.search(
            r"days left:\s*(Unlimited|[-]?\d+)", auth_output, re.IGNORECASE
        )

        if serial_key_match and days_left_match:
            auth_serial = serial_key_match.group(1).strip()
            auth_serial = self._format_serial_key(auth_serial)
            days_raw = days_left_match.group(1).strip()

            try:
                is_unlimited = days_raw.lower() == "unlimited"
            except AttributeError:
                is_unlimited = False

            if is_unlimited:
                logger.warning(
                    f"Serial key {auth_serial} returned 'Unlimited' — "
                    f"treating as invalid."
                )
                days_left = -1
            else:
                try:
                    days_left = int(days_raw)
                except ValueError:
                    days_left = -1

            logger.info(
                f"Authorization info: serial key = {auth_serial}, "
                f"days left = {days_left}"
            )

            if days_left < 0:

                try:
                    output, success, deactivated_keys = self.deactivate_serial()
                    if success:
                        logger.info(
                            "Deactivation command -W executed successfully "
                            f"for {auth_serial}"
                        )
                    else:
                        logger.warning(
                            "Deactivation command -W failed for " f"{auth_serial}"
                        )
                except Exception as e:
                    logger.error(f"Error during CLI deactivation (-W): {e}")

                try:
                    expired_serial = SerialHistory.objects.filter(
                        serial_key=auth_serial
                    ).first()
                    if expired_serial:
                        expired_serial.status = "expired"
                        expired_serial.save()
                        logger.warning(
                            f"Serial key {auth_serial} is expired or invalid. "
                            f"Days left: {days_left}"
                        )
                    else:
                        logger.warning(
                            f"Serial key {auth_serial} not found in " f"SerialHistory."
                        )
                except Exception as e:
                    logger.error(
                        "Error updating expiration status for serial key "
                        f"{auth_serial}: {e}"
                    )

                try:
                    config = Config.objects.first()
                    if config:
                        config.is_activated = False
                        config.save()
                        logger.warning(
                            "Config updated: is_activated=False due to "
                            "expired or invalid serial key."
                        )
                except Exception as e:
                    logger.error(f"Error updating config for expired serial key: {e}")
            else:
                try:
                    serial_obj = SerialHistory.objects.filter(
                        serial_key=auth_serial
                    ).first()
                    if serial_obj:
                        serial_obj.remaining_days = days_left
                        serial_obj.save()
                        logger.info(
                            "Updated remaining days for serial key "
                            f"{auth_serial} to {days_left}."
                        )
                    else:
                        logger.warning(
                            f"Serial key {auth_serial} not found in " f"SerialHistory."
                        )
                except Exception as e:
                    logger.error(
                        "Error updating remaining days for serial key "
                        f"{auth_serial}: {e}"
                    )

        history: List[SerialHistory] = SerialHistory.objects.all()
        logger.info(f"Serial history: {list(history)}")
        return [
            {
                "serial_key": entry.serial_key,
                "status": entry.status,
                "remaining_days": entry.remaining_days,
            }
            for entry in history
        ]

    def _parse_serial_history(self, output: str) -> List[str]:
        """
        Parses the output of the KelVPN -H command and returns a list of serial
        keys.

        :param output: The standard output from the KelVPNCli -H command.
        :return: List of serial keys.
        """
        serial_keys: List[str] = []
        lines: List[str] = output.splitlines()

        # Find the line containing "Serial History size"
        size_line: Optional[str] = next(
            (line for line in lines if "Serial History size" in line), None
        )
        if size_line:
            try:
                size: int = int(
                    re.search(r"Serial History size: (\d+)", size_line).group(1)
                )
            except (AttributeError, ValueError):
                size = 0
        else:
            size = 0

        for line in lines:
            match = re.match(r"\d{2} : ([A-Z0-9-]{19})", line)
            if match:
                serial_keys.append(match.group(1))

        return serial_keys

    def deactivate_serial(self) -> Tuple[str, bool, Optional[List[str]]]:
        """
        Deactivates all active serial key(s) by resetting them using
        KelVPNCli (-W) and updates the serial history.

        If the output indicates success (or that there are no activated keys),
        all active serial keys are marked as deactivated and the configuration is
        updated (is_activated=False).

        :return: Tuple containing the command output, success flag, and a list of
                 deactivated serial key strings (or None if none).
        """
        cmd_reset_serial: List[str] = [self.cli_path, "-W"]
        output: str = self.execute_command(cmd_reset_serial)
        logger.debug(f"Deactivation command output: {output}")

        if ("Success!" in output and "Resetting the serial key." in output) or (
            "There are no activated keys." in output
        ):
            active_serials = list(SerialHistory.objects.filter(status="active"))
            deactivated_keys = []
            if active_serials:
                for serial in active_serials:
                    serial.status = "deactivated"
                    serial.save()
                    deactivated_keys.append(serial.serial_key)
                    logger.info(
                        f"Serial key {serial.serial_key} deactivated " f"successfully."
                    )
            else:
                logger.warning("No active serial keys found during deactivation.")

            self.update_config(is_connected=False, is_activated=False)
            return output, True, deactivated_keys if deactivated_keys else None

        logger.error(f"Failed to deactivate serial key. Command output: {output}")
        return output, False, None

    def get_server_pings(self) -> List[Dict[str, Any]]:
        servers_output: str = self.execute_command([self.cli_path, "-L"])
        actual_servers = {}
        server_line_regex = re.compile(r"\s*(\d{2})\s*\|\s*(.+?)\s*$")
        for line in servers_output.splitlines():
            match = server_line_regex.match(line)
            if match:
                server_id, server_name = (
                    match.group(1).strip(),
                    match.group(2).strip(),
                )
                actual_servers[server_name] = server_id

        ping_output: str = self.execute_command([self.cli_path, "-L", "-R"])
        ping_line_regex = re.compile(r"\s*\d{2}\s*\|\s*(\d+)\s*\|\s*(.+?)\s*$")

        servers_ping = []
        for line in ping_output.splitlines():
            match = ping_line_regex.match(line)
            if match:
                ping = int(match.group(1).strip())
                server_name = match.group(2).strip()

                actual_server_id = actual_servers.get(server_name)
                if actual_server_id:
                    servers_ping.append(
                        {
                            "server_id": actual_server_id,
                            "name": server_name,
                            "ping": ping,
                            "selected": False,
                        }
                    )

        with transaction.atomic():
            config = Config.objects.select_related("selected_server").first()
            existing_selected_server_id = (
                config.selected_server.server_id
                if config and config.selected_server
                else None
            )

            for server in servers_ping:
                is_selected = server["server_id"] == existing_selected_server_id

                server_obj, _ = Server.objects.update_or_create(
                    server_id=server["server_id"],
                    defaults={
                        "name": server["name"],
                        "ping": server["ping"],
                        "selected": is_selected,
                    },
                )

                if is_selected:
                    server["selected"] = True

            if (
                not config.selected_server
                or not Server.objects.filter(
                    server_id=existing_selected_server_id
                ).exists()
            ):
                Server.objects.filter(selected=True).update(selected=False)
                best_server = min(servers_ping, key=lambda s: s["ping"], default=None)
                if best_server:
                    selected_server = Server.objects.get(
                        server_id=best_server["server_id"]
                    )
                    selected_server.selected = True
                    selected_server.save()
                    config.selected_server = selected_server
                    config.save()

        return sorted(servers_ping, key=lambda s: s["ping"])

    def _parse_server_ping_list(self, output: str) -> List[Dict[str, Any]]:
        """
        Parses the output of the KelVPNCli -L -R command and returns a list of
        servers with pings.

        :param output: The standard output from the KelVPNCli -L -R command.
        :return: List of dictionaries containing server details.
        """
        servers_ping: List[Dict[str, Any]] = []
        lines: List[str] = output.splitlines()

        server_ping_line_regex = re.compile(
            r"^\s*(\d{2})\s*\|\s*(\d{1,4})\s*\|\s*(.+?)\s*$"
        )

        for line in lines:
            match = server_ping_line_regex.match(line)
            if match:
                server_id: str = match.group(1).strip()
                ping_str: str = match.group(2).strip()
                server_name: str = match.group(3).strip()

                try:
                    ping: int = int(ping_str)
                except ValueError:
                    ping = -1

                if server_id and server_name:
                    servers_ping.append(
                        {
                            "id": server_id,
                            "ping": ping,
                            "name": server_name,
                            "selected": False,
                        }
                    )
                    logger.debug(
                        f"Parsed server with ping: ID={server_id}, "
                        f"Ping={ping}, Name={server_name}"
                    )

        logger.info(f"Total servers with ping parsed: {len(servers_ping)}")
        return servers_ping

    def get_location(self) -> str:
        try:
            response = subprocess.check_output(
                ["curl", "-s", "http://ip-api.com/json/"]
            )
            json_data = json.loads(response.decode("utf-8"))
            country = json_data.get("country", "-")
            city = json_data.get("city", "")
            return f"{country}, {city}" if city else country
        except Exception as e:
            logger.warning(f"Failed to fetch location info: {e}")
            return "-"

    def update_config(
        self,
        is_connected: bool,
        is_activated: Optional[bool] = None,
        server_name: Optional[str] = None,
    ) -> None:
        """
        Updates the config's connection and activation status.

        :param is_connected: Boolean indicating if the service is connected.
        :param is_activated: Optional boolean indicating if the service is
                             activated.
        """
        config: Optional[Config] = Config.objects.first()
        location = self.get_location()

        if config:
            config.is_connected = is_connected
            config.location = location
            if is_activated is not None:
                config.is_activated = is_activated
            config.save()

        try:
            with open(self.config_path, "r") as file:
                lines = file.readlines()

            new_lines = []
            found_connected = False
            found_server = False

            for line in lines:
                if line.startswith("KELVPN_CONNECTED="):
                    new_lines.append(
                        f'KELVPN_CONNECTED={"true" if is_connected else "false"}\n'
                    )
                    found_connected = True
                elif line.startswith("KELVPN_SERVER=") and server_name:
                    new_lines.append(f'KELVPN_SERVER="{server_name}"\n')
                    found_server = True
                else:
                    new_lines.append(line)

            if not found_connected:
                new_lines.append(
                    f'KELVPN_CONNECTED={"true" if is_connected else "false"}\n'
                )
            if not found_server and server_name:
                new_lines.append(f'KELVPN_SERVER="{server_name}"\n')

            with open(self.config_path, "w") as file:
                file.writelines(new_lines)

            logger.info(
                f"Updated configuration: KELVPN_CONNECTED={'true' if is_connected else 'false'}, KELVPN_SERVER={server_name}"
            )

        except Exception as e:
            logger.error(f"Failed to update configuration: {e}")

    def _get_ip_addresses(self):
        # result = subprocess.run(["nmcli", "device", "show"], capture_output=True, text=True)
        # addresses = re.findall(r"IP4\.ADDRESS\[\d+\]:\s+([\d.]+)", result.stdout)
        try:
            external_ip = (
                subprocess.check_output(["curl", "-s", "ifconfig.me"]).decode().strip()
            )
            # logger.debug(f"ip {external_ip}")
            return external_ip

        except subprocess.CalledProcessError as e:
            logger.debug(f"Error retrieving external IP: {e}")
            return None

    def _get_gateway(self):
        """
        Получает IP-адрес шлюза с приоритетом на tun0 (VPN), затем eth0 (Ethernet) и, наконец, wlan0 (Wi-Fi).
        """
        result = subprocess.run(
            ["nmcli", "device", "show"], capture_output=True, text=True
        )
        output = result.stdout

        # Определяем приоритет интерфейсов
        priority_interfaces = ["tun0", "eth0", "wlan0"]

        for interface in priority_interfaces:
            if f"GENERAL.DEVICE:                         {interface}" in output:
                match = re.search(
                    rf"{interface}\n(?:.*\n)*?IP4\.GATEWAY:\s+([\d.]+)", output
                )
                if match:
                    return match.group(1)

        return "-"

    def _get_default_interface(self):
        result = subprocess.run(
            ["ip", "route", "show", "default"], capture_output=True, text=True
        )
        match = re.search(r"default via [\d.]+ dev (\S+)", result.stdout)
        if match:
            return match.group(1)
        return None

    def _get_dns_for_interface(self, iface):
        result = subprocess.run(
            ["resolvectl", "dns", iface], capture_output=True, text=True
        )
        match = re.search(rf"\({iface}\):\s+([\d\.]+)", result.stdout)
        if match:
            return [match.group(1)]
        return []

    def _get_current_active_dns(self):
        default_iface = self._get_default_interface()
        if not default_iface:
            return "-"
        dns_servers = self._get_dns_for_interface(default_iface)
        return dns_servers[0] if dns_servers else "-"

    def update_network_config(self):
        """
        Updates network configuration and remaining days of VPN usage.
        """
        ip_addresses = self._get_ip_addresses()
        gateway = self._get_gateway()
        active_dns = self._get_current_active_dns()

        logger.info(
            f"Updating network config: ip={ip_addresses}, gateway={gateway}, dns={active_dns}"
        )

        # Get or create the configuration record
        config, created = Config.objects.get_or_create(id=1)

        config.ip = ip_addresses if ip_addresses else "-"
        config.gateway = gateway
        config.dns = active_dns if active_dns else "-"

        config.save()
        logger.info("Network configuration updated successfully.")


class WifiManager:
    CONFIG_PATHS = settings.HOSTAPD_CONFIG_PATHS
    # Добавляем глобальный объект блокировки для координации перезапусков
    _restart_timer = None
    _restart_lock = threading.Lock()
    _pending_restart = set()

    @staticmethod
    def get_lock_path(interface: str) -> str:
        return f"/tmp/wifi_manager_{interface}.lock"

    @staticmethod
    def start_hostapd(interface: str) -> None:
        """
        Starts the hostapd service for the given interface.
        """
        subprocess.run(["systemctl", "start", f"hostapd@{interface}"], check=True)

    @staticmethod
    def stop_hostapd(interface: str) -> None:
        """
        Stops the hostapd service for the given interface.
        """
        subprocess.run(["systemctl", "stop", f"hostapd@{interface}"], check=True)

    @staticmethod
    def restart_hostapd(interface: str = None) -> None:
        """
        Restarts the hostapd service for the given interface.
        If interface is None, restarts the main hostapd service.
        """
        if interface:
            subprocess.run(["systemctl", "restart", f"hostapd@{interface}"], check=True)
        else:
            # Restart the main hostapd service
            subprocess.run(["systemctl", "restart", "hostapd"], check=True)
            logger.info("Restarted main hostapd service")

    @classmethod
    def schedule_hostapd_restart(cls, interface: str) -> None:
        """
        Schedules a hostapd restart to be executed after a short delay.
        This allows multiple configuration changes to be batched together
        before triggering a single restart.
        """
        with cls._restart_lock:
            # Add this interface to the pending restart set
            cls._pending_restart.add(interface)

            # If a timer is already running, we don't need to create a new one
            if cls._restart_timer is not None:
                logger.debug(f"Added {interface} to pending hostapd restart queue")
                return

            # Create a timer to execute the restart after a delay
            def do_restart():
                with cls._restart_lock:
                    # Clear the timer reference first
                    cls._restart_timer = None

                    # Get the set of interfaces that need restarting
                    interfaces = cls._pending_restart.copy()
                    cls._pending_restart.clear()

                    logger.info(
                        f"Executing delayed restart for interfaces: {interfaces}"
                    )

                    if len(interfaces) > 1:
                        # If we have multiple interfaces, just restart the main service once
                        cls.restart_hostapd()
                    elif len(interfaces) == 1:
                        # If only one interface, restart just that one
                        cls.restart_hostapd(next(iter(interfaces)))

            # Start a timer to execute restart after 500ms
            timer = threading.Timer(0.5, do_restart)
            timer.daemon = True
            timer.start()
            cls._restart_timer = timer
            logger.debug(
                f"Scheduled hostapd restart in 500ms for interface {interface}"
            )

    @staticmethod
    def toggle_hostapd(interface: str, enable: bool) -> None:
        """
        Toggles the hostapd service based on the `enable` flag.
        """
        if enable:
            WifiManager.start_hostapd(interface)
        else:
            WifiManager.stop_hostapd(interface)

    @staticmethod
    def read_configs() -> list[dict]:
        """
        Reads configurations from all hostapd config files.
        Returns a list of configuration dictionaries.
        """
        configs = []
        for interface, config_path in WifiManager.CONFIG_PATHS.items():
            config = {"interface": interface, "config_file": config_path}
            try:
                with open(config_path, "r") as file:
                    for line in file:
                        line = line.strip()
                        if line and not line.startswith("#") and "=" in line:
                            key, value = line.split("=", 1)
                            config[key.strip()] = value.strip()
                configs.append(config)
            except FileNotFoundError:
                logger.error(f"Configuration file {config_path} not found.")
        return configs

    @staticmethod
    def determine_band(config: dict) -> str:
        """
        Determines the Wi-Fi band based on the configuration.
        """
        hw_mode = config.get("hw_mode", "g")
        channel = int(config.get("channel", 0))

        if hw_mode == "g" and 1 <= channel <= 14:
            return "2.4GHz"
        if hw_mode == "a" and 36 <= channel <= 165:
            return "5GHz"
        if hw_mode == "6" and 1 <= channel <= 233:
            return "6GHz"
        return "Unknown"

    @staticmethod
    def update_wifi_status() -> None:
        """
        Updates the WiFiStatus records based on configurations.
        """
        configs = WifiManager.read_configs()
        for config in configs:
            try:
                WiFiStatus.objects.update_or_create(
                    interface=config["interface"],
                    defaults={
                        "band": WifiManager.determine_band(config),
                        "ssid": config.get("ssid", "-"),
                        "channel": config.get("channel", "-"),
                        "hw_mode": config.get("hw_mode", "-"),
                        "transmit_power": config.get("transmit_power", 0),
                        "enabled": WifiManager.check_service_status(
                            config["interface"]
                        ),
                        "wpa_passphrase": config.get("wpa_passphrase", "-"),
                    },
                )
            except Exception as e:
                logger.error(f"Failed to update Wi-Fi status: {e}")

    @staticmethod
    def check_service_status(interface: str) -> bool:
        """
        Checks whether the hostapd service for the given interface is active.
        """
        result = subprocess.run(
            ["systemctl", "is-active", f"hostapd@{interface}"],
            stdout=subprocess.PIPE,
            text=True,
        )
        return result.stdout.strip() == "active"

    @staticmethod
    def update_network_config(interface: str, **kwargs) -> None:
        """
        Updates Wi-Fi network configuration parameters in the hostapd
        config file. Only provided kwargs are updated.
        """
        config_path = WifiManager.CONFIG_PATHS.get(interface)
        if not config_path:
            raise FileNotFoundError(f"Configuration file for {interface} not found.")

        config_data = {}
        if os.path.exists(config_path):
            with open(config_path, "r") as file:
                for line in file:
                    line = line.strip()
                    if line and not line.startswith("#") and "=" in line:
                        key, value = line.split("=", 1)
                        config_data[key.strip()] = value.strip()

        for key, value in kwargs.items():
            if value is not None:
                config_data[key] = str(value)

        with open(config_path, "w") as file:
            for key, value in config_data.items():
                file.write(f"{key}={value}\n")

        # Вместо немедленного перезапуска планируем отложенный перезапуск
        WifiManager.schedule_hostapd_restart(interface)

    @classmethod
    def safe_update_wifi_status(cls) -> None:
        """
        Safe (atomic) update of Wi-Fi statuses with a file lock.
        """
        try:
            with FileLock(LOCK_PATH, timeout=LOCK_TIMEOUT):
                cls.update_wifi_status()
        except Timeout:
            logger.error("Failed to acquire lock to update Wi-Fi statuses.")
            raise Exception("Could not acquire lock to update Wi-Fi status.")

    @classmethod
    def safe_update_network_config(cls, interface: str, **kwargs):
        lock_path = cls.get_lock_path(interface)
        try:
            with FileLock(lock_path, timeout=LOCK_TIMEOUT):
                logger.debug(f"Lock acquired for interface {interface}")
                cls.update_network_config(interface, **kwargs)
        except Timeout:
            logger.error(f"Timeout while waiting lock on {interface}")
            raise Exception(f"Could not acquire lock to update config {interface}")

    @classmethod
    def safe_toggle_hostapd(cls, interface: str, enable: bool):
        lock_path = cls.get_lock_path(interface)
        try:
            with FileLock(lock_path, timeout=LOCK_TIMEOUT):
                logger.debug(f"Lock acquired for toggling interface {interface}")
                if enable:
                    cls.start_hostapd(interface)
                else:
                    cls.stop_hostapd(interface)
        except Timeout:
            logger.error(f"Timeout while toggling hostapd on {interface}")
            raise Exception(f"Could not acquire lock to toggle hostapd {interface}")


class UpdateService:
    """
    Package update service:

    * perform_update()        - only kelvpn packages
    * perform_system_update() - all system packages (apt upgrade)
    * perform_custom_update() - user-specified package list from frontend

    The process is launched in a transient-unit via **systemd-run**,
    so restarting the backend won't interrupt the installation.
    """

    KELVPN_PACKAGES = ["kelvpn", "kelvpn-admin"]
    LOG_DIR = "/opt/kelvpn-web-admin/backend/logs"

    # ------------------------------- helpers -------------------------------
    @staticmethod
    def _get_package_versions(pkg_name: str) -> tuple[str, str]:
        try:
            logger.debug(f"Checking versions for package: {pkg_name}")
            res = subprocess.run(
                ["apt-cache", "policy", pkg_name],
                text=True,
                check=True,
                capture_output=True,
            )
            installed, candidate = "unknown", "unknown"
            for line in res.stdout.splitlines():
                if "Installed:" in line:
                    installed = line.split("Installed:")[-1].strip()
                elif "Candidate:" in line:
                    candidate = line.split("Candidate:")[-1].strip()
            logger.debug(
                f"Package '{pkg_name}': installed={installed}, candidate={candidate}"
            )
            return installed, candidate
        except subprocess.CalledProcessError as e:
            logger.error(f"apt-cache policy failed for {pkg_name}: {e.stderr}")
            return "unknown", "unknown"

    @staticmethod
    def _spawn_systemd_unit(args: List[str], log_suffix: str) -> Tuple[str, str]:
        """
        Launches *args* in a transient-unit. Returns (unit_name, log_path).
        """
        ts = datetime.now().strftime("%Y%m%d_%H%M%S")
        unit = f"kelvpn-{log_suffix}-{ts}"
        log_path = f"{UpdateService.LOG_DIR}/{log_suffix}_{ts}.log"

        cmd = [
            "systemd-run",
            "--unit",
            unit,
            "--description",
            f"KelVPN update ({log_suffix})",
            "--collect",
            "--property",
            f"StandardOutput=file:{log_path}",
            "--property",
            f"StandardError=file:{log_path}",
        ] + args

        logger.debug("Launch update unit: %s", " ".join(cmd))
        subprocess.run(cmd, check=True)
        return unit, log_path

    @staticmethod
    def _unit_finished(unit: str) -> Tuple[bool, bool]:
        """
        Checks the status of the systemd unit.

        Returns:
            Tuple[bool, bool]: (finished, success)
        """
        try:
            out = subprocess.check_output(
                ["systemctl", "show", "--property=ActiveState,Result", unit], text=True
            )
            data = dict(kv.split("=", 1) for kv in out.strip().splitlines())
            finished = data.get("ActiveState") in ("inactive", "failed")
            success = data.get("Result") == "success"
            return finished, success
        except subprocess.CalledProcessError:
            # unit was removed by --collect option -> consider it finished successfully
            return True, True

    @staticmethod
    def _start_update(unit_args: List[str], suffix: str) -> Tuple[bool, str]:
        st, _ = UpdateStatus.objects.get_or_create(id=1)
        if st.status == "IN_PROGRESS":
            return False, "Update already in progress"

        unit, log_path = UpdateService._spawn_systemd_unit(unit_args, suffix)

        st.status = "IN_PROGRESS"
        st.unit_name = unit
        st.log_file = log_path
        st.started_at = datetime.now()
        st.completed_at = None
        st.save()

        return True, f"{suffix.replace('-', ' ').title()} started"

    # ------------------------------- public API -------------------------------

    @staticmethod
    def perform_update() -> Tuple[bool, str]:
        return UpdateService._start_update(
            ["/usr/local/bin/kelvpn-self-update"], "kelvpn-update"
        )

    @staticmethod
    def perform_system_update() -> Tuple[bool, str]:
        return UpdateService._start_update(
            ["bash", "-lc", "apt-get update -y && apt-get upgrade -y"],
            "system-update",
        )

    @staticmethod
    def perform_custom_update(packages: List[str]) -> Tuple[bool, str]:
        pkg_list = " ".join(packages)
        return UpdateService._start_update(
            ["bash", "-lc", f"apt-get update -y && apt-get install -y {pkg_list}"],
            "custom-update",
        )

    # ------------------------------- check / status -------------------------------

    @staticmethod
    def check_updates() -> UpdateCheckResponse:
        logger.info("Starting update check...")

        kelvpn_ver, kelvpn_candidate = UpdateService._get_package_versions("kelvpn")
        backend_ver, backend_candidate = UpdateService._get_package_versions(
            "kelvpn-admin"
        )
        ap_ver, ap_candidate = UpdateService._get_package_versions("kelvpn-ap")

        # Refresh APT index
        logger.debug("Refreshing apt cache with 'apt-get update'")
        subprocess.run(
            ["apt-get", "update", "-y"],
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        updates: List[UpdateItem] = []
        kelvpn_updates: List[UpdateItem] = []
        system_updates: List[UpdateItem] = []

        for name, current, candidate in [
            ("kelvpn", kelvpn_ver, kelvpn_candidate),
            ("kelvpn-admin", backend_ver, backend_candidate),
            ("kelvpn-ap", ap_ver, ap_candidate),
        ]:
            if current != "unknown" and candidate != "unknown" and current != candidate:
                logger.info(f"Update available for {name}: {current} -> {candidate}")
                item = UpdateItem(package=name, current=current, candidate=candidate)
                updates.append(item)
                kelvpn_updates.append(item)
            else:
                logger.debug(f"No update for {name}: {current} == {candidate}")

        # Get system updates for visibility
        try:
            res = subprocess.run(
                ["apt", "list", "--upgradable"],
                text=True,
                check=True,
                capture_output=True,
            )
            for ln in res.stdout.splitlines()[1:]:
                if "upgradable from:" not in ln:
                    continue
                logger.debug(f"System update: {ln}")
                parts = ln.split()
                if len(parts) >= 2:
                    pkg = parts[0].split("/")[0]
                    candidate = parts[1]
                    current = ln.split("upgradable from:")[-1].strip(" ]")
                    if pkg not in {"kelvpn", "kelvpn-admin", "kelvpn-ap"}:
                        system_updates.append(
                            UpdateItem(
                                package=pkg, current=current, candidate=candidate
                            )
                        )
        except subprocess.CalledProcessError as e:
            logger.warning("Failed to list system updates: %s", e.stderr)

        updates_available = bool(kelvpn_updates)

        return UpdateCheckResponse(
            updates_available=updates_available,
            kelvpn_version=kelvpn_ver,
            backend_version=backend_ver,
            kelvpn_ap_version=ap_ver,
            updates=kelvpn_updates,
            system_updates=system_updates,
        )

    @staticmethod
    def get_update_status():
        st, _ = UpdateStatus.objects.get_or_create(id=1)

        logger.debug(
            f"Checking update status. Current DB state: "
            f"status={st.status}, unit={st.unit_name}, started_at={st.started_at}"
        )

        if st.status == "IN_PROGRESS":
            if not st.unit_name:
                logger.warning(
                    "Detected IN_PROGRESS status without unit_name. Marking as FAILED."
                )
                st.status = "FAILED"
                st.completed_at = datetime.now()
                st.save()
            else:
                finished, success = UpdateService._unit_finished(st.unit_name)
                logger.debug(
                    f"Systemd unit check: unit_name={st.unit_name}, finished={finished}, success={success}"
                )
                if finished:
                    st.status = "COMPLETED" if success else "FAILED"
                    st.completed_at = datetime.now()
                    st.save()
                    logger.info(
                        f"Update unit {st.unit_name} finished. "
                        f"Marked as {'COMPLETED' if success else 'FAILED'}."
                    )

        return {
            "status": st.status,
            "log_file": (
                f"logs/{os.path.basename(st.log_file)}" if st.log_file else None
            ),
            "started_at": st.started_at,
            "completed_at": st.completed_at,
        }


class NetworkManager:
    WAN_INTERFACE = "wan0"

    @staticmethod
    def get_network_settings():
        settings = {
            "mac_address": NetworkManager.get_mac_address(),
            "ip_address": NetworkManager.get_ip_address(),
            "gateway": NetworkManager.get_default_gateway(),
            "ip_mode": NetworkManager.get_ip_mode(),
        }
        return settings

    @staticmethod
    def get_mac_address():
        result = subprocess.run(
            ["cat", f"/sys/class/net/{NetworkManager.WAN_INTERFACE}/address"],
            stdout=subprocess.PIPE,
            text=True,
        )
        return result.stdout.strip()

    @staticmethod
    def set_mac_address(mac_address: str):
        if not re.match(r"^([0-9A-Fa-f]{2}:){5}[0-9A-Fa-f]{2}$", mac_address):
            raise ValueError("Invalid MAC address format.")

        subprocess.run(["ip", "link", "set", NetworkManager.WAN_INTERFACE, "down"])
        subprocess.run(
            ["ip", "link", "set", NetworkManager.WAN_INTERFACE, "address", mac_address]
        )
        subprocess.run(["ip", "link", "set", NetworkManager.WAN_INTERFACE, "up"])

    @staticmethod
    def get_ip_address():
        result = subprocess.run(
            ["ip", "-4", "addr", "show", NetworkManager.WAN_INTERFACE],
            stdout=subprocess.PIPE,
            text=True,
        )
        match = re.search(r"inet (\d+\.\d+\.\d+\.\d+)", result.stdout)
        return match.group(1) if match else "Not Assigned"

    @staticmethod
    def get_default_gateway():
        result = subprocess.run(
            ["ip", "route", "show", "default"],
            stdout=subprocess.PIPE,
            text=True,
        )
        match = re.search(r"default via (\d+\.\d+\.\d+\.\d+)", result.stdout)
        return match.group(1) if match else "Not Assigned"

    @staticmethod
    def get_ip_mode():
        dhcp_file = f"/var/lib/dhcp/dhclient.{NetworkManager.WAN_INTERFACE}.leases"
        if os.path.exists(dhcp_file):
            return "Dynamic"
        return "Static"

    @staticmethod
    def set_ip_config(static: bool, ip_address=None, netmask=None, gateway=None):
        if static:
            if not all([ip_address, netmask, gateway]):
                raise ValueError("IP address, netmask, and gateway must be set.")
            subprocess.run(["ip", "addr", "flush", "dev", NetworkManager.WAN_INTERFACE])
            subprocess.run(
                [
                    "ip",
                    "addr",
                    "add",
                    f"{ip_address}/{netmask}",
                    "dev",
                    NetworkManager.WAN_INTERFACE,
                ]
            )
            subprocess.run(["ip", "route", "add", "default", "via", gateway])
        else:
            subprocess.run(["dhclient", NetworkManager.WAN_INTERFACE])


@register_socket_handlers
class SocketMonitorService:
    """
    Service for monitoring the KelVPN socket and processing messages.
    Uses TCP socket communication with JSON protocol.
    Handlers are registered using the @command_handler decorator.
    """

    _instance = None
    _monitor_task = None

    @classmethod
    def get_instance(cls):
        """Singleton pattern implementation"""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.host = "127.0.0.1"
        self.port = 22143
        self.greeting_timeout = 15.0
        self.response_timeout = 15.0
        self.reconnect_delay = 5
        self.running = False
        self.socket = None
        self.monitor_task = None

    async def start_monitoring(self):
        """
        Start the socket monitoring service if it's not already running.
        This method is non-blocking and returns immediately.
        """
        if self.running:
            logger.info("Socket monitor is already running")
            return

        self.running = True

        # Start the main monitor task
        if self.__class__._monitor_task is None or self.__class__._monitor_task.done():
            self.__class__._monitor_task = asyncio.create_task(self._monitor_socket())
            logger.info(f"Started socket monitoring task for {self.host}:{self.port}")

    async def stop_monitoring(self):
        """
        Stop the socket monitoring service.
        """
        if not self.running:
            return

        self.running = False

        # Cancel monitor task
        if self.__class__._monitor_task and not self.__class__._monitor_task.done():
            self.__class__._monitor_task.cancel()
            try:
                await self.__class__._monitor_task
            except asyncio.CancelledError:
                pass
            self.__class__._monitor_task = None

        # Close the connection
        if self.socket:
            self._close_socket()

        logger.info("Socket monitoring stopped")

    def _create_socket(self):
        """
        Create a new socket connection.
        """
        import socket

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((self.host, self.port))
            sock.settimeout(0)
            logger.info(f"Connected to socket at {self.host}:{self.port}")
            return sock
        except (socket.error, ConnectionRefusedError) as e:
            logger.warning(
                f"Could not connect to socket at {self.host}:{self.port}: {e}"
            )
            return None

    def _close_socket(self):
        """
        Close the socket connection safely.
        """
        if self.socket:
            try:
                self.socket.close()
                logger.info("Socket connection closed")
            except Exception as e:
                logger.error(f"Error closing socket: {e}")
            finally:
                self.socket = None

    def _recv_data(self, timeout=0.5):
        """
        Receive data from socket with timeout.
        """
        if not self.socket:
            return b""

        import socket
        import select

        buf = []
        try:
            # Use select for non-blocking recv with timeout
            ready = select.select([self.socket], [], [], timeout)
            if ready[0]:
                while True:
                    try:
                        # Non-blocking recv
                        chunk = self.socket.recv(4096)
                        if not chunk:
                            break
                        buf.append(chunk)
                    except socket.error:
                        # No more data available
                        break
        except Exception as e:
            logger.error(f"Error receiving data: {e}")

        return b"".join(buf)

    def _parse_json_stream(self, raw_bytes):
        """
        Extract all JSON objects from a byte stream.
        """
        if not raw_bytes:
            return []

        try:
            s = raw_bytes.decode("utf-8", "replace")
            import json

            dec = json.JSONDecoder()
            idx = 0
            out = []
            n = len(s)
            while idx < n:
                # Skip leading whitespace
                while idx < n and s[idx].isspace():
                    idx += 1
                if idx >= n:
                    break
                try:
                    obj, length = dec.raw_decode(s[idx:])
                    out.append(obj)
                    idx += length
                except json.JSONDecodeError:
                    # No more JSON to parse
                    break
            return out
        except Exception as e:
            logger.error(f"Error parsing JSON stream: {e}")
            return []

    async def _monitor_socket(self):
        """
        Main monitoring loop. Connects to the socket and reads messages.
        If connection is lost, it will attempt to reconnect after a delay.
        """
        try:
            while self.running:
                if not self.socket:
                    self.socket = self._create_socket()
                    if not self.socket:
                        logger.warning(
                            "Socket connection failed, retrying after delay..."
                        )
                        await asyncio.sleep(self.reconnect_delay)
                        continue

                    # Получаем initial greeting
                    greeting_data = self._recv_data(self.greeting_timeout)
                    greeting_messages = self._parse_json_stream(greeting_data)
                    for msg in greeting_messages:
                        self._process_message(msg)

                try:
                    data = self._recv_data(0.5)

                    if data:
                        messages = self._parse_json_stream(data)
                        for msg in messages:
                            self._process_message(msg)
                    else:
                        # Проверяем, не закрыт ли сокет сервером
                        if self.socket:
                            import select

                            r, _, _ = select.select([self.socket], [], [], 0)
                            if r:
                                chunk = self.socket.recv(1024)
                                if not chunk:
                                    raise ConnectionResetError("Socket disconnected")

                except (
                    ConnectionResetError,
                    ConnectionAbortedError,
                    BrokenPipeError,
                ) as e:
                    logger.warning(f"Connection lost ({e}), reconnecting...")
                    self._close_socket()

                except Exception as e:
                    logger.error(f"Unexpected error in socket monitor: {e}")
                    self._close_socket()

                await asyncio.sleep(0.5)

        except asyncio.CancelledError:
            logger.info("Socket monitoring task cancelled")
            raise
        except Exception as e:
            logger.error(f"Unexpected error in socket monitor: {e}")
            if self.running:
                # Restart monitoring after a delay
                asyncio.create_task(self._delayed_restart())

    async def _delayed_restart(self):
        """Restart monitoring after a delay to prevent rapid restarts on error"""
        await asyncio.sleep(10)
        if self.running:
            self.__class__._monitor_task = asyncio.create_task(self._monitor_socket())

    def _process_message(self, message):
        """
        Process a message from the socket by finding and calling the registered handler.
        """
        command = message.get("command", "")
        result = message.get("result", None)
        logger.info(f"← Got `{command}`: {json.dumps(result or message)}")

        # Get the unbound method from the class handlers dict
        handler_method = self.__class__._handlers.get(command)
        if handler_method:
            try:
                # Create a bound method associated with the instance 'self'
                bound_handler = handler_method.__get__(self, self.__class__)

                # Check if the original method was async
                if asyncio.iscoroutinefunction(handler_method):
                    # Schedule the async bound method call
                    asyncio.create_task(bound_handler(message))
                else:
                    # Call the sync bound method directly
                    bound_handler(message)
            except Exception as e:
                logger.error(f"Error in handler `{command}`: {e}", exc_info=True)
        else:
            logger.debug(f"No handler registered for command: {command}")

    @command_handler("get states")
    def _on_states(self, message):
        """
        The get states message handler:
        updates the session_state, stream_state, tunnel_state, and is_connected fields.
        """
        result = message.get("result", {})
        states = result.get("states", {})

        session = states.get("session", "false")
        stream = states.get("stream", "false")
        tunnel = states.get("tunnel", "false")

        asyncio.create_task(self._save_states(session, stream, tunnel))

    @sync_to_async
    def _save_states(self, session, stream, tunnel):
        cfg, _ = Config.objects.get_or_create(pk=1)
        cfg.session_state = session
        cfg.stream_state = stream
        cfg.tunnel_state = tunnel
        cfg.save()
        cfg.update_connection_flag()
        logger.info(
            f"Config updated: session={session}, stream={stream}, tunnel={tunnel}, is_connected={cfg.is_connected}"
        )

    @command_handler("send server data")
    async def _on_send_server_data(self, message):
        """
        Handles 'send server data' messages.
        Updates the selected_server in Config if the received server differs.
        """
        result = message.get("result", {})
        server_name = result.get("name")

        if not server_name:
            logger.warning("Received 'send server data' without server name.")
            return

        # Schedule the database update asynchronously
        asyncio.create_task(self._update_selected_server(server_name))

    @sync_to_async
    def _update_selected_server(self, server_name: str):
        """Finds server by name and updates Config if it's not the selected one."""
        try:
            cfg, _ = Config.objects.select_related("selected_server").get_or_create(
                pk=1
            )
            target_server = Server.objects.filter(name=server_name).first()

            if not target_server:
                logger.warning(f"Server '{server_name}' not found in database.")
                return

            current_selected_server_name = (
                cfg.selected_server.name if cfg.selected_server else None
            )

            if current_selected_server_name != server_name:
                cfg.selected_server = target_server
                cfg.save()
                logger.info(
                    f"Updated selected server in Config to: {server_name} (ID: {target_server.server_id})"
                )
            else:
                logger.debug(f"Server '{server_name}' is already the selected server.")

        except Exception as e:
            logger.error(
                f"Error updating selected server to '{server_name}': {e}", exc_info=True
            )

    @command_handler("request reset license key")
    async def _on_request_reset_license_key(self, message):
        """
        Handles 'request reset license key' messages.
        If successful, updates Config and SerialHistory.
        """
        result = message.get("result", {})
        reply = result.get("reset_reply", "")

        if "reset successfully" in reply.lower():
            logger.info("License key reset request successful. Updating database.")
            asyncio.create_task(self._reset_activation_status())
        else:
            logger.warning(
                f"License key reset request failed or reply not recognized: {reply}"
            )

    @sync_to_async
    def _reset_activation_status(self):
        """Sets config.is_activated to False and deactivates non-expired serials."""
        try:
            with transaction.atomic():
                # Update Config
                cfg, _ = Config.objects.get_or_create(pk=1)
                if cfg.is_activated:
                    cfg.is_activated = False
                    cfg.save()
                    logger.info("Set Config.is_activated to False.")

                # Update SerialHistory
                updated_count = SerialHistory.objects.exclude(status="expired").update(
                    status="deactivated",
                )
                if updated_count > 0:
                    logger.info(
                        f"Deactivated {updated_count} serial key(s) in SerialHistory."
                    )
                else:
                    logger.info("No active/deactivated serial keys found to update.")

        except Exception as e:
            logger.error(
                f"Error resetting activation status in database: {e}", exc_info=True
            )

    def _parse_days_left_string(self, days_str: str) -> Optional[int]:
        """Parses strings like '362 days left' into an integer."""
        if not days_str:
            return None
        match = re.search(r"(\d+)\s+days left", days_str, re.IGNORECASE)
        if match:
            try:
                return int(match.group(1))
            except (ValueError, IndexError):
                logger.warning(f"Could not parse days from string: {days_str}")
        return None

    def _parse_license_time(self, timestamp_str: str) -> Optional[datetime]:
        """Parses a Unix timestamp string into a timezone-aware datetime object."""
        if not timestamp_str:
            return None
        try:
            timestamp = int(timestamp_str)
            # Assume the timestamp is UTC
            return datetime.fromtimestamp(timestamp, tz=dt_timezone.utc)
        except ValueError:
            logger.warning(f"Could not parse timestamp string: {timestamp_str}")
        return None

    @command_handler("data local")
    async def _on_data_local(self, message):
        """
        Handles 'data local' messages, specifically the 'serialKeyData' part.
        Updates or creates SerialHistory and updates Config.
        """
        result = message.get("result", {})
        data = result.get("data", {})
        serial_data = data.get("serialKeyData", {})

        if not isinstance(serial_data, dict):
            logger.debug("'data local' received without valid 'serialKeyData'.")
            # logger.debug(f"'data local' message: {message}")
            return

        serial_key_raw = serial_data.get("serialKey")
        is_activated = serial_data.get("isActivated")
        days_left_str = serial_data.get("daysLeftString")
        license_time_str = serial_data.get("lisenseTime")

        if not serial_key_raw:
            logger.warning("Received 'serialKeyData' without 'serialKey'.")
            return

        # Parse data before passing to the database function
        serial_key_formatted = self._format_serial_key(serial_key_raw)
        remaining_days = self._parse_days_left_string(days_left_str)
        expired_date = self._parse_license_time(license_time_str)
        
        # Schedule the database update
        asyncio.create_task(
            self._update_serial_and_config(
                serial_key=serial_key_formatted,
                status=is_activated,
                remaining_days=remaining_days,
                expired_date=expired_date,
            )
        )

    @sync_to_async
    def _update_serial_and_config(
        self, serial_key: str, status: bool, remaining_days: Optional[int], expired_date: Optional[datetime] = None
    ):
        """Updates SerialHistory and Config based on data local message."""
        try:
            with transaction.atomic():
                # First, get or create the SerialHistory with minimal defaults
                serial_obj, created = SerialHistory.objects.get_or_create(
                    serial_key=serial_key,
                    defaults={
                        "status": "active" if status else "deactivated",
                        "remaining_days": -1,
                    },
                )

                # Now update the object with all the details
                serial_obj.status = "active" if status else "deactivated"

                if remaining_days is not None:
                    serial_obj.remaining_days = remaining_days

                if expired_date and expired_date < datetime.now(dt_timezone.utc):
                    # License has expired
                    serial_obj.status = "expired"
                    serial_obj.remaining_days = 0
                    logger.info(
                        f"Serial key {serial_key} has expired. Setting SerialHistory.status to 'expired' and SerialHistory.remaining_days to 0."
                    )
                serial_obj.save()

                if created:
                    logger.info(
                        f"Created new SerialHistory for key {serial_key}. Status: {serial_obj.status}, Days Left: {remaining_days}"
                    )
                else:
                    logger.info(
                        f"Updated SerialHistory for key {serial_key}. Status: {serial_obj.status}, Days Left: {remaining_days}"
                    )

                # Update Config
                cfg, _ = Config.objects.get_or_create(pk=1)
                config_updated = False
                if cfg.is_activated != status:
                    cfg.is_activated = status
                    config_updated = True

                if status == "active" and remaining_days is not None:
                    if cfg.remaining_days != remaining_days:
                        cfg.remaining_days = remaining_days
                        config_updated = True

                if config_updated:
                    cfg.save()
                    logger.info(
                        f"Updated Config: is_activated={cfg.is_activated}, remaining_days={cfg.remaining_days}"
                    )

        except Exception as e:
            logger.error(
                f"Error processing 'data local' for serial {serial_key}: {e}",
                exc_info=True,
            )

    def send_command(self, command, params):
        """
        Send a command to the socket.

        Args:
            command: Command name
            params: Command parameters dict

        Returns:
            List of response objects
        """
        if not self.socket:
            logger.error("Cannot send command: Socket not connected")
            return []

        try:
            cmd = {"command": command, "params": params}
            payload = json.dumps(cmd) + "\n\n"
            self.socket.sendall(payload.encode("utf-8"))
            logger.debug(f"Sent command: {command}, params: {params}")

            # Wait for response
            response_data = self._recv_data(self.response_timeout)
            responses = self._parse_json_stream(response_data)

            logger.debug(f"Received {len(responses)} responses")
            return responses

        except Exception as e:
            logger.error(f"Error sending command {command}: {e}")
            self._close_socket()
            return []

    def get_servers_list(self):
        """
        Get list of available VPN servers.

        Returns:
            List of server objects or empty list on error
        """
        if not self.socket:
            self.socket = self._create_socket()
            if not self.socket:
                logger.error("Cannot get servers: Socket not connected")
                return []

        responses = self.send_command("get servers list", {})

        for response in responses:
            cmd = response.get("command") == "get servers list"
            has_result = "result" in response
            has_servers = has_result and "servers" in response["result"]
            if cmd and has_servers:
                return response["result"]["servers"]

        return []

    def get_connection_state(self):
        """
        Get the current connection state.

        Returns:
            Connection state dict or empty dict on error
        """
        if not self.socket:
            self.socket = self._create_socket()
            if not self.socket:
                logger.error("Cannot get connection state: Socket not connected")
                return {}

        responses = self.send_command("get states", {})

        for response in responses:
            if response.get("command") == "get states" and "result" in response:
                return response["result"]

        return {}

    @staticmethod
    def _format_serial_key(serial: str) -> str:
        """
        Formats a serial key into the database format (XXXX-XXXX-XXXX-XXXX).
        If the serial key already contains hyphens or its length is not 16, returns it unchanged.
        """
        if not serial:
            return ""
        serial = serial.replace("-", "")
        if len(serial) != 16:
            return serial
        return f"{serial[0:4]}-{serial[4:8]}-{serial[8:12]}-{serial[12:16]}"

    @command_handler("get servers list")
    async def _on_get_servers_list(self, message):
        """
        Handles 'get servers list' messages.
        Updates the ip_address field for servers found in the database.
        """
        result = message.get("result", {})
        servers = result.get("servers")

        if not isinstance(servers, list):
            logger.warning("'get servers list' received without a valid 'servers' list.")
            return

        logger.debug(f"Received {len(servers)} server entries from 'get servers list'.")
        for server_data in servers:
            if isinstance(server_data, dict) and "Name" in server_data and "Address" in server_data:
                asyncio.create_task(self._update_server_ip(server_data))
            else:
                logger.warning(f"Skipping invalid server data entry: {server_data}")

    @sync_to_async
    def _update_server_ip(self, server_data: Dict[str, Any]):
        """Updates the IP address for a server found by name."""
        server_name = server_data.get("Name")
        server_ip = server_data.get("Address")

        if not server_name or not server_ip:
            logger.warning(f"Missing Name or Address in server data: {server_data}")
            return

        try:
            server_obj = Server.objects.filter(name=server_name).first()
            if server_obj:
                if server_obj.ip_address != server_ip:
                    server_obj.ip_address = server_ip
                    server_obj.save()
                    logger.info(f"Updated IP address for server '{server_name}' to {server_ip}.")
                else:
                    # Optional: Log if the IP is already up-to-date
                    logger.debug(f"IP address for server '{server_name}' is already {server_ip}.")
            else:
                # Optional: Log if the server from the socket is not found in DB
                logger.debug(f"Server '{server_name}' from socket not found in DB, skipping IP update.")

        except Exception as e:
            logger.error(
                f"Error updating IP address for server '{server_name}': {e}",
                exc_info=True
            )


MONITORED_INTERFACES = {"tun0", "wan0", "wlan0"}
DNS_TEST_DOMAIN = "google.com"
RECONNECT_COMMAND = ["systemctl", "restart", "KelVPNService.service"]


class NetworkMonitor:
    _instance = None
    _monitor_task = None

    @classmethod
    def get_instance(cls):
        """Singleton pattern implementation"""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        if self.__class__._instance:
            raise Exception("Singleton NetworkMonitor already instantiated")
        self.ipr = None
        self.running = False
        self._monitor_task = None
        self.reconnect_delay = 10  # Seconds to wait before restarting loop on error
        logger.info("NetworkMonitor initialized.")

    async def start_monitoring(self):
        """
        Start the network monitoring service if it's not already running.
        Manages the main monitoring loop and handles restarts on errors.
        """
        if self.running:
            logger.info("Network monitor is already running.")
            return

        self.running = True
        logger.info("Starting Network Monitor...")

        # Loop to manage the core monitoring task and restart it if it fails
        while self.running:
            try:
                # Create and await the main monitoring task
                self._monitor_task = asyncio.create_task(self._run_loop())
                await self._monitor_task
                # If the task finishes without error (unexpected for an infinite loop)
                if self.running:
                    logger.warning(
                        "Network monitor loop finished unexpectedly. Restarting..."
                    )

            except asyncio.CancelledError:
                logger.info("Network monitor task explicitly cancelled.")
                self.running = False  # Ensure loop terminates if cancelled externally
                break
            except Exception as e:
                logger.error(
                    f"Network monitor loop error: {e}. Restarting after delay.",
                    exc_info=True,
                )
            finally:
                # Cleanup IPRoute instance before potential restart or final exit
                if self.ipr:
                    try:
                        await asyncio.get_running_loop().run_in_executor(
                            None, self.ipr.close
                        )
                        logger.info("IPRoute instance closed cleanly.")
                    except Exception as close_err:
                        logger.error(f"Error closing IPRoute: {close_err}")
                    finally:
                        self.ipr = None

                if self.running:  # Only sleep if we intend to restart
                    logger.info(
                        f"Waiting {self.reconnect_delay} seconds before restarting monitor..."
                    )
                    await asyncio.sleep(self.reconnect_delay)

        logger.info("Network Monitor has stopped.")
        self.running = False  # Final state check

    async def stop_monitoring(self):
        """
        Stop the network monitoring service gracefully.
        """
        if not self.running:
            logger.info("Network monitor is not running.")
            return

        logger.info("Stopping Network Monitor...")
        self.running = False  # Signal the loop to stop

        # Cancel the main task
        if self._monitor_task and not self._monitor_task.done():
            self._monitor_task.cancel()
            try:
                await self._monitor_task  # Wait for cancellation to complete
            except asyncio.CancelledError:
                logger.info("Network monitor task successfully cancelled.")
            except Exception as e:
                logger.error(
                    f"Error awaiting network monitor task cancellation: {e}",
                    exc_info=True,
                )

        # Final cleanup is handled in the start_monitoring finally block
        self._monitor_task = None
        logger.info("Network Monitor stopping process complete.")

    async def _run_loop(self):
        """
        Core monitoring loop that listens for network events.
        Must be managed by start_monitoring.
        """
        loop = asyncio.get_running_loop()
        try:
            # Create IPRoute instance using run_in_executor to avoid event loop conflicts
            self.ipr = await loop.run_in_executor(None, lambda: IPRoute(netns=None))
            
            # bind is potentially blocking
            await loop.run_in_executor(None, self.ipr.bind)
            logger.info("NetworkMonitor bound to netlink successfully.")

            while self.running:
                # get() is blocking, run in executor
                events = await loop.run_in_executor(None, self.ipr.get)
                for event in events:
                    # Process event non-blockingly
                    asyncio.create_task(self._handle_event(event))

                # Brief sleep to yield control and prevent busy-waiting if get() returns empty frequently
                await asyncio.sleep(0.1)

        except Exception as e:
            logger.error(f"Error in NetworkMonitor run loop: {e}", exc_info=True)
            # Reraise to be caught by the outer management loop in start_monitoring
            raise
        finally:
            # Cleanup is handled by the caller (start_monitoring)
            logger.debug("Exiting _run_loop.")

    async def _handle_event(self, event):
        """Handle a single network event."""
        # RTM_NEWLINK=16, RTM_DELLINK=17
        if event.get("header", {}).get("type") not in {16, 17}:
            return

        attrs = dict(event.get("attrs", []))
        iface_name = attrs.get("IFLA_IFNAME")

        if iface_name not in MONITORED_INTERFACES:
            return

        event_type = event.get("header", {}).get("type")
        logger.warning(f"Detected network event ({event_type}) on {iface_name}.")

        # Schedule diagnosis non-blockingly
        asyncio.create_task(self._diagnose_and_repair(iface_name))

    async def _diagnose_and_repair(self, iface_triggered: str):
        """Diagnose network state and potentially trigger repairs."""
        logger.info(
            f"Running network diagnostics triggered by event on {iface_triggered}..."
        )
        # Short delay to allow network state to stabilize after event
        await asyncio.sleep(1.0)

        ip_ok = await self._check_ip_address()
        route_ok = await self._check_default_route()
        dns_ok = await self._check_dns_resolution()

        if ip_ok and route_ok and dns_ok:
            logger.info("✅ Network status appears OK after check.")
            return

        logger.error(
            f"⚠️ Network problem detected! IP OK: {ip_ok}, Route OK: {route_ok}, DNS OK: {dns_ok}"
        )

        # Placeholder for repair logic (e.g., restarting VPN service)
        # logger.info("Attempting to reconnect KelVPNService...")
        # try:
        #     # Run shell command asynchronously if needed, or use KelVPNCliService
        #     # process = await asyncio.create_subprocess_exec(*RECONNECT_COMMAND)
        #     # await process.wait()
        #     # if process.returncode == 0:
        #     #      logger.info("✅ KelVPNService restarted successfully.")
        #     # else:
        #     #      logger.error(f"❌ KelVPNService restart failed with code {process.returncode}.")
        #     pass # Currently disabled
        # except Exception as e:
        #     logger.error(f"❌ Failed to run KelVPNService restart command: {e}", exc_info=True)

    async def _check_ip_address(self) -> bool:
        """Check if monitored interfaces have an IP address."""
        if not self.ipr:
            logger.warning("Cannot check IP address, IPRoute not initialized.")
            return False
        loop = asyncio.get_running_loop()
        try:
            for iface in MONITORED_INTERFACES:
                # link_lookup might block
                links = await loop.run_in_executor(
                    None, getattr, self.ipr, "link_lookup", {"ifname": iface}
                )
                if not links:
                    continue
                # get_addr might block
                addrs = await loop.run_in_executor(
                    None, getattr, self.ipr, "get_addr", links[0]
                )
                # Check for valid IPv4/IPv6 addresses
                if any(
                    a.get("family") in {socket.AF_INET, socket.AF_INET6} for a in addrs
                ):
                    logger.debug(f"Interface {iface} has IP address assigned.")
                    return True
            logger.warning("No monitored interface has an IP address.")
            return False
        except Exception as e:
            logger.error(f"Error checking IP addresses: {e}", exc_info=True)
            return False

    async def _check_default_route(self) -> bool:
        """Check if a default route exists."""
        if not self.ipr:
            logger.warning("Cannot check default route, IPRoute not initialized.")
            return False
        loop = asyncio.get_running_loop()
        try:
            # get_routes might block
            routes = await loop.run_in_executor(
                None, getattr, self.ipr, "get_routes", socket.AF_INET
            )
            for route in routes:
                # Check for destination 0.0.0.0/0
                if route.get("dst_len") == 0:
                    gateway = dict(route.get("attrs", [])).get("RTA_GATEWAY")
                    logger.debug(f"Default route found via {gateway}.")
                    return True
            logger.warning("No default route found.")
            return False
        except Exception as e:
            logger.error(f"Error checking default route: {e}", exc_info=True)
            return False

    async def _check_dns_resolution(self) -> bool:
        """Check if DNS resolution works for a test domain."""
        loop = asyncio.get_running_loop()
        try:
            await loop.getaddrinfo(DNS_TEST_DOMAIN, None, family=socket.AF_INET)
            logger.debug(f"DNS resolution for {DNS_TEST_DOMAIN} successful.")
            return True
        except socket.gaierror:
            logger.warning(f"DNS resolution failed for {DNS_TEST_DOMAIN}.")
            return False
        except Exception as e:
            logger.error(f"Error during DNS resolution check: {e}", exc_info=True)
            return False
